from sklearn.linear_model import Lasso
from sklearn.preprocessing import MinMaxScaler, PolynomialFeatures
import numpy as np
import pandas as pd
from symbal.utils import get_equation
import copy


class LassoEnsemble:

    def __init__(self, alpha_range=(1e-4, 1e3), num_models=20, scaler=MinMaxScaler(),
                 polynomial=PolynomialFeatures(degree=2), max_iter=50000, progress=True, equation_file=None,
                 selection='AIC'):

        self.best_model = None
        alpha_list = np.geomspace(alpha_range[0], alpha_range[1], num_models)
        self.all_models = [Lasso(alpha=alpha, max_iter=max_iter) for alpha in alpha_list]
        self.models = None
        self.scaler = scaler
        self.polynomial = polynomial
        self.column_list = None
        self.equations_ = None
        self.eq_ = None
        self.equation_file = equation_file
        self.progress = progress
        self.selection = selection

    def fit(self, x: pd.DataFrame, y: pd.Series):

        self.column_list = list(x.columns)
        x = np.array(x)
        y = np.array(y)

        x_poly = self.polynomial.fit_transform(x)
        if self.scaler is not None:
            x_poly_scale = self.scaler.fit_transform(x_poly)
        else:
            x_poly_scale = x_poly

        complexities, losses, equations = [[], [], []]

        for model in self.all_models:

            model.fit(x_poly_scale, y)
            complexities.append(np.sum(model.coef_.astype(bool)))
            losses.append(np.sqrt(np.mean((y - model.predict(x_poly_scale)) ** 2)))
            equations.append(get_equation(self.column_list, model, self.scaler, self.polynomial))

        self.eq_ = pd.DataFrame(np.array([losses, complexities]).T,
                                columns=['loss', 'complexity'])
        idmax = list(self.eq_.groupby('complexity')['loss'].idxmax())
        self.eq_ = self.eq_.iloc[idmax, :].reset_index(drop=True)
        self.models = copy.deepcopy(self.all_models)
        self.models = [model for i, model in enumerate(self.models) if i in idmax]
        equations = [equation for i, equation in enumerate(equations) if i in idmax]

        self.eq_['nLoss'] = (self.eq_['loss'] - np.min(self.eq_['loss'])) / np.ptp(self.eq_['loss'])
        self.eq_['nComp'] = (self.eq_['complexity'] - np.min(self.eq_['complexity'])) / np.ptp(self.eq_['complexity'])

        if self.selection == 'AIC':
            self.eq_['metric'] = len(x) * np.log(self.eq_['loss'] / len(x)) + 2 * self.eq_['complexity']
        elif self.selection == 'AICc':
            self.eq_['metric'] = (len(x) * np.log(self.eq_['loss'] / len(x)) + 2 * self.eq_['complexity'] +
                                  (2 * self.eq_['complexity'] * (self.eq_['complexity']
                                                                 + 1)/(len(x) - self.eq_['complexity'] - 1)))
        elif self.selection == 'nAIC':
            self.eq_['metric'] = len(x) * np.log(self.eq_['nLoss'] / len(x)) + 2 * self.eq_['nComp']
        elif self.selection == 'nAICc':
            self.eq_['metric'] = (len(x) * np.log(self.eq_['nLoss'] / len(x)) + 2 * self.eq_['nComp'] +
                                  (2 * self.eq_['nComp'] * (self.eq_['nComp'] + 1)/(len(x) - self.eq_['nComp'] - 1)))
        elif self.selection == 'BIC':
            self.eq_['metric'] = len(x) * np.log(self.eq_['loss'] / len(x)) + self.eq_['complexity'] * np.log(len(x))
        elif self.selection == 'nBIC':
            self.eq_['metric'] = len(x) * np.log(self.eq_['nLoss'] / len(x)) + self.eq_['nComp'] * np.log(len(x))

        self.eq_.loc[0, 'metric'] = 0
        self.eq_.loc[0, 'pareto'] = 0
        self.eq_.replace([np.inf, -np.inf, np.nan], 0, inplace=True)

        for i in range(self.eq_.shape[0]):
            if i > 0:

                idmin = self.eq_.loc[:i-1, :]['loss'].idxmin()
                self.eq_.loc[i, 'pareto'] = ((self.eq_.loc[i, 'metric'] - self.eq_.loc[idmin, 'metric']) / (
                        self.eq_.loc[i, 'complexity'] - self.eq_.loc[idmin, 'complexity']))

        self.eq_.loc[0:1, 'pareto'] = 0

        self.eq_['nPar'] = (self.eq_['pareto'] - np.mean(self.eq_['pareto'])) / np.std(self.eq_['pareto'])
        self.eq_['score'] = np.exp((np.min(self.eq_['nPar']) - self.eq_['nPar']) / 2)

        self.eq_ = self.eq_[self.eq_['pareto'] <= 0]
        self.models = [model for i, model in enumerate(self.models) if i in list(self.eq_.index)]
        equations = [equation for i, equation in enumerate(equations) if i in list(self.eq_.index)]
        self.eq_ = self.eq_.reset_index(drop=True)

        best_index = self.eq_['pareto'].idxmin()
        self.eq_['model'] = self.models
        self.eq_['equation'] = equations
        self.best_model = self.eq_.loc[best_index, 'model']

        self.equations_ = self.eq_

        if self.progress:
            self.equations_.to_csv(f'{self.equation_file}')

    def predict(self, x: pd.DataFrame, num=None):

        x = x[self.column_list]
        x = np.array(x)

        x_poly = self.polynomial.transform(x)
        if self.scaler is not None:
            x_poly_scale = self.scaler.transform(x_poly)
        else:
            x_poly_scale = x_poly

        if num is None:
            y_pred = self.best_model.predict(x_poly_scale)
        else:
            y_pred = self.equations_.loc[num, 'model'].predict(x_poly_scale)
        return y_pred




import numpy as np
from scipy.stats import qmc
import pandas as pd
import itertools as it
import random


class RandomEngine(qmc.QMCEngine):

    def __init__(self, d, seed=None):
        super().__init__(d=d, seed=seed)

    def _random(self, n=1, *, workers=1):
        return self.rng.random((n, self.d))

    def reset(self):
        super().__init__(d=self.d, seed=self.rng_seed)
        return self

    def fast_forward(self, n):
        self.random(n)
        return self


class TestFunction:

    def __init__(self, function, min_val, max_val, fract_extr=0.1, num_inter=30, num_extr=30,
                 method_inter='MC', method_extr='MC', method_cand='FF', cand_grid_size=30, initial_size=30,
                 seed=None, noise_level=0.1):

        self.candidates = None
        self.initial_set = None
        self.extrapolation_testset = None
        self.interpolation_testset = None
        self.min_vals = None
        self.max_vals = None
        self.function = function

        if isinstance(min_val, list):
            self.min_vals = np.array(min_val)
        elif isinstance(min_val, np.ndarray):
            self.min_vals = min_val
        else:
            raise TypeError('min_val should be of type list or np.ndarray.')

        if isinstance(max_val, list):
            self.max_vals = np.array(max_val)
        elif isinstance(max_val, np.ndarray):
            self.max_vals = max_val
        else:
            raise TypeError('max_val should be of type list or np.ndarray.')

        assert self.min_vals.shape == self.max_vals.shape, 'min_vals and max_vals are different sizes'

        self.init_extrap_set(function, fract_extr, num_extr, method_extr, seed, noise_level)
        self.init_interp_set(function, fract_extr, num_inter, method_inter, seed, noise_level)
        self.init_candidate_set(function, fract_extr, method_cand, cand_grid_size, initial_size, noise_level)

    def init_extrap_set(self, function, fract_extr, num_extr, method_extr, seed, noise_level):

        if method_extr == 'MC':
            pass  # TODO - add different extrapolation sampling methods

        n_extrs = 0
        self.extrapolation_testset = np.empty((num_extr, self.min_vals.shape[0]), dtype=float)
        engine = RandomEngine(d=self.min_vals.shape[0], seed=seed)

        lower_bound = self.min_vals + (fract_extr / 2) * (self.max_vals - self.min_vals)
        upper_bound = self.max_vals - (fract_extr / 2) * (self.max_vals - self.min_vals)

        while n_extrs < num_extr:

            sample = engine.random(1).flatten()
            sample = (self.max_vals - self.min_vals) * sample + self.min_vals

            bool_extr = np.any((sample < lower_bound) | (sample > upper_bound))
            if bool_extr:
                self.extrapolation_testset[n_extrs, :] = sample
                n_extrs += 1

        variables = [f'x{i+1}' for i in range(self.min_vals.shape[0])]
        extr_df = pd.DataFrame(self.extrapolation_testset, columns=variables)
        extr_df['output'] = extr_df.eval(function)
        extr_df['output'] = extr_df['output'] + np.random.normal(scale=noise_level*np.std(extr_df['output']),
                                                                 size=(num_extr,))
        self.extrapolation_testset = extr_df

    def init_interp_set(self, function, fract_extr, num_inter, method_inter, seed, noise_level):

        if method_inter == 'MC':
            pass  # TODO - add different interpolation sampling methods

        n_inter = 0
        self.interpolation_testset = np.empty((num_inter, self.min_vals.shape[0]), dtype=float)
        engine = RandomEngine(d=self.min_vals.shape[0], seed=seed)

        lower_bound = self.min_vals + (fract_extr / 2) * (self.max_vals - self.min_vals)
        upper_bound = self.max_vals - (fract_extr / 2) * (self.max_vals - self.min_vals)

        while n_inter < num_inter:

            sample = engine.random(1).flatten()
            sample = (self.max_vals - self.min_vals) * sample + self.min_vals

            bool_extr = np.any((sample < lower_bound) | (sample > upper_bound))
            if not bool_extr:
                self.interpolation_testset[n_inter, :] = sample
                n_inter += 1

        variables = [f'x{i+1}' for i in range(self.min_vals.shape[0])]
        inter_df = pd.DataFrame(self.interpolation_testset, columns=variables)
        inter_df['output'] = inter_df.eval(function)
        inter_df['output'] = inter_df['output'] + np.random.normal(scale=noise_level*np.std(inter_df['output']),
                                                                   size=(num_inter,))
        self.interpolation_testset = inter_df

    def init_candidate_set(self, function, fract_extr, method_cand, cand_grid_size, initial_size, noise_level):

        if method_cand == 'FF':
            pass  # TODO - add different candidate sampling methods

        lower_bound = self.min_vals + (fract_extr / 2) * (self.max_vals - self.min_vals)
        upper_bound = self.max_vals - (fract_extr / 2) * (self.max_vals - self.min_vals)

        comb_list = []

        for lower, upper in zip(lower_bound, upper_bound):
            comb_list.append(np.linspace(lower, upper, cand_grid_size))

        candidate_array = np.array(list(it.product(*comb_list)))

        ind_initial = random.sample(range(len(candidate_array)), k=initial_size)
        initial_array = candidate_array[ind_initial, :]
        candidate_array = np.delete(candidate_array, ind_initial, axis=0)

        variables = [f'x{i + 1}' for i in range(self.min_vals.shape[0])]
        self.candidates = pd.DataFrame(candidate_array, columns=variables)
        self.candidates['output'] = self.candidates.eval(function)
        self.candidates['output'] += np.random.normal(scale=noise_level*np.std(self.candidates['output']),
                                                      size=(len(self.candidates['output']),))

        self.initial_set = pd.DataFrame(initial_array, columns=variables)
        self.initial_set['output'] = self.initial_set.eval(function)
        self.initial_set['output'] += np.random.normal(scale=noise_level*np.std(self.initial_set['output']),
                                                       size=(len(self.initial_set['output']),))




import pickle


class Capsule:

    def __init__(self,
                 dataframe=None,
                 output_name=None,
                 x_units=None,
                 y_units=None,
                 test_df=None,
                 groupby=None):

        self.dataframe = dataframe
        self.output_name = output_name
        self.x_units = x_units
        self.y_units = y_units
        self.test_df = test_df
        self.groupby = groupby

    def store_capsule(self, filename):

        with open(filename, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)


def open_capsule(filename):

    with open(filename, 'rb') as handle:
        capsule = pickle.load(handle)

    return capsule




import copy

from symbal.utils import get_gradient, get_curvature, get_uncertainties
from symbal.utils import get_fim, get_optimal_krr, get_optimal_gpr
import numpy as np
import scipy as sp
from scipy.spatial.distance import cdist
from sklearn.preprocessing import StandardScaler
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern, RationalQuadratic
from scipy.interpolate import Rbf
from operator import itemgetter


def objective(cand_df, exist_df, pysr_model, acquisition, batch_config):

    x_cand = cand_df.drop('output', axis=1)
    x_exist = exist_df.drop('output', axis=1)

    function_dict = {
        'gradient':     {'func': _gradient,     'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'curvature':    {'func': _curvature,    'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'grad1':        {'func': _grad1,        'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'curv1':        {'func': _curv1,        'params': {'x_cand',  'pysr_model', 'batch_config'}, 'scale': True},
        'uncertainty':  {'func': _uncertainty,  'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'certainty':    {'func': _certainty,    'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'Aopt':         {'func': _aopt,         'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'Dopt':         {'func': _dopt,         'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'Eopt':         {'func': _eopt,         'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'boundary':     {'func': _boundary,     'params': ['x_cand',  'pysr_model', 'batch_config'], 'scale': True},
        'distance':     {'func': _distance,     'params': ['x_cand',  'x_exist',    'batch_config'], 'scale': False},
        'proximity':    {'func': _proximity,    'params': ['x_cand',  'x_exist',    'batch_config'], 'scale': False},
        'density':      {'func': _density,      'params': ['x_cand',  'x_exist',    'batch_config'], 'scale': False},
        'sparsity':     {'func': _sparsity,     'params': ['x_cand',  'x_exist',    'batch_config'], 'scale': False},
        'gaussian_unc': {'func': _gaussian_unc, 'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'know_grad':    {'func': _know_grad,    'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'leaveoneout':  {'func': _leaveoneout,  'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'LOOA':         {'func': _looa,         'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'LOOM':         {'func': _loom,         'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'GUGS':         {'func': _gugs,         'params': ['cand_df', 'exist_df',   'batch_config'], 'scale': True},
        'random':       {'func': _random,       'params': ['x_cand'],                                'scale': False},
        'rand1':        {'func': _rand1,        'params': ['x_cand'],                                'scale': False},
    }
    param_dict = {'cand_df': cand_df, 'exist_df': exist_df, 'x_cand': x_cand, 'x_exist': x_exist,
                  'pysr_model': pysr_model, 'batch_config': batch_config}

    objective_array = np.zeros((len(x_cand)))

    for method in acquisition:

        if method not in function_dict:
            raise KeyError(f'{method} does not exist as an acquisition strategy. Available methods:'
                           f' {list(function_dict.keys())}')

        if len(function_dict[method]['params']) == 1:
            values = function_dict[method]['func'](itemgetter(*function_dict[method]['params'])(param_dict))
        else:
            values = function_dict[method]['func'](*itemgetter(*function_dict[method]['params'])(param_dict))
        values = _scale_objective(values, batch_config) if function_dict[method]['scale'] else values
        objective_array += acquisition[method] * values

    if 'debug' in batch_config:
        if batch_config['debug']:

            print_string = f'max: {np.max(objective_array)}, min: {np.min(objective_array)}, '
            print_string += f'avg: {np.mean(objective_array)}, std: {np.std(objective_array)}'
            print(print_string)

    return objective_array


def _gradient(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8

    gradient_array = np.empty((len(x_cand), len(pysr_model.equations_['equation'])))
    for j, _ in enumerate(pysr_model.equations_['equation']):
        gradient_array[:, j] = get_gradient(x_cand, pysr_model, num=j, difference=difference)

    if 'score_reg' in batch_config:
        if batch_config['score_reg']:
            scores = np.array(pysr_model.equations_['score'])
            gradient_array = gradient_array * scores

    gradients = np.sum(np.abs(gradient_array), axis=1)
    return gradients


def _curvature(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8

    curvature_array = np.empty((len(x_cand), len(pysr_model.equations_['equation'])))
    for j, _ in enumerate(pysr_model.equations_['equation']):
        curvature_array[:, j] = get_curvature(x_cand, pysr_model, num=j, difference=difference)

    if 'score_reg' in batch_config:
        if batch_config['score_reg']:
            scores = np.array(pysr_model.equations_['score'])
            curvature_array = curvature_array * scores

    curvatures = np.sum(np.abs(curvature_array), axis=1)
    return curvatures


def _distance(x_cand, x_exist, batch_config):

    distance_metric = batch_config['distance_metric'] if 'distance_metric' in batch_config else 'euclidean'

    cand_array = np.array(x_cand)
    exist_array = np.array(x_exist)
    cand_norm = (cand_array - np.min(cand_array, axis=0)) / np.ptp(cand_array, axis=0)
    exist_norm = (exist_array - np.min(cand_array, axis=0)) / np.ptp(cand_array, axis=0)

    dist_array = cdist(cand_norm, exist_norm, metric=distance_metric)
    dist_vector = np.min(dist_array, axis=1)
    return dist_vector


def _proximity(x_cand, x_exist, batch_config):
    dist_vector = _distance(x_cand, x_exist, batch_config)
    return -dist_vector


def _density(x_cand, x_exist, batch_config):

    distance_metric = batch_config['distance_metric'] if 'distance_metric' in batch_config else 'euclidean'

    cand_array = np.array(x_cand)
    exist_array = np.array(x_exist)
    cand_norm = (cand_array - np.min(cand_array, axis=0)) / np.ptp(cand_array, axis=0)
    exist_norm = (exist_array - np.min(cand_array, axis=0)) / np.ptp(cand_array, axis=0)

    dist_array = cdist(cand_norm, exist_norm, metric=distance_metric)
    dens_vector = np.mean(dist_array, axis=1)
    return dens_vector


def _sparsity(x_cand, x_exist, batch_config):
    dens_vector = _density(x_cand, x_exist, batch_config)
    return -dens_vector


def _uncertainty(x_cand, pysr_model, batch_config):

    uncertainty_array = get_uncertainties(x_cand, pysr_model)

    if 'score_reg' in batch_config:
        if batch_config['score_reg']:
            scores = np.array(pysr_model.equations_['score'])
            uncertainty_array = uncertainty_array * scores

    uncertainties = np.sum(np.abs(uncertainty_array), axis=1)
    return uncertainties


def _certainty(x_cand, pysr_model, batch_config):
    uncertainties = _uncertainty(x_cand, pysr_model, batch_config)
    return -uncertainties


def _random(x_cand):
    random_array = np.random.uniform(size=(len(x_cand),))
    return random_array


def _rand1(x_cand):

    random_array = np.random.normal(size=(len(x_cand),))
    random_array = (random_array - np.min(random_array)) / np.ptp(random_array)
    return random_array


def _grad1(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8
    gradient_array = get_gradient(x_cand, pysr_model, difference=difference)
    gradients = np.abs(gradient_array)
    return gradients


def _curv1(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8
    curvature_array = get_curvature(x_cand, pysr_model, difference=difference)
    curvatures = np.abs(curvature_array)
    return curvatures


def _gaussian_unc(cand_df, exist_df, batch_config):
    _, y_cand_std = __gaussian_fit(cand_df, exist_df, batch_config)
    return y_cand_std


def _know_grad(cand_df, exist_df, batch_config):

    y_cand_mean, y_cand_std = __gaussian_fit(cand_df, exist_df, batch_config)

    z = np.zeros(shape=y_cand_mean.shape)

    y_exist_max = np.max(exist_df['output'])
    z[y_cand_std != 0.] = (y_cand_mean[y_cand_std != 0.] - y_exist_max) / y_cand_std[y_cand_std != 0.]

    cdf = sp.stats.norm.cdf(z)
    pdf = sp.stats.norm.pdf(z)

    gradients = np.zeros(shape=y_cand_mean.shape)
    gradients[z != 0.] = y_cand_std[z != 0.] * cdf[z != 0.] + pdf[z != 0.]

    return gradients


def _aopt(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8
    traces = np.zeros((len(x_cand),))

    if 'subset' in batch_config:
        x_cand = x_cand.sample(batch_config['subset'])

    for i, cand in enumerate(list(x_cand.index)):
        fim = get_fim(x_cand.loc[cand, :], pysr_model, single=True, difference=difference)
        traces[cand] = np.trace(fim)

    return traces


def _dopt(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8
    dets = np.zeros((len(x_cand),))

    if 'subset' in batch_config:
        x_cand = x_cand.sample(batch_config['subset'])

    for i, cand in enumerate(list(x_cand.index)):
        fim = get_fim(x_cand.loc[cand, :], pysr_model, single=True, difference=difference)
        dets[cand] = np.linalg.det(fim)

    return dets


def _eopt(x_cand, pysr_model, batch_config):

    difference = batch_config['difference'] if 'difference' in batch_config else 1e-8
    mineigs = np.zeros((len(x_cand),))

    if 'subset' in batch_config:
        x_cand = x_cand.sample(batch_config['subset'])

    for i, cand in enumerate(list(x_cand.index)):
        fim = get_fim(x_cand.loc[cand, :], pysr_model, single=True, difference=difference)
        mineigs[cand] = np.min(np.linalg.eigvals(fim))

    return mineigs


def _boundary(x_cand, pysr_model, batch_config):

    power = batch_config['power'] if 'power' in batch_config else 1

    y_pred = pysr_model.predict(x_cand)

    if 'scale_pred' in batch_config:
        if batch_config['scale_pred']:
            y_pred = y_pred / np.ptp(y_pred)
    else:
        y_pred = y_pred / np.ptp(y_pred)

    values = np.exp(-np.abs(y_pred) ** power)
    return values


def _leaveoneout(cand_df, exist_df, batch_config):

    cand_dfc = copy.deepcopy(cand_df)
    exist_dfc = copy.deepcopy(exist_df)

    alpha_number = batch_config['alpha_number'] if 'alpha_number' in batch_config else 10

    if 'krr_param_grid' in batch_config:
        krr_param_grid = batch_config['krr_param_grid']
    else:
        krr_param_grid = {
            'alpha': [alpha for alpha in np.geomspace(1e-4, 1e4, alpha_number)],
            'kernel': ['linear', 'poly', 'rbf', 'sigmoid'],
            'degree': [2, 3],
        }

    Scaler = batch_config['scaler'] if 'scaler' in batch_config else StandardScaler
    grid_rounds = batch_config['grid_rounds'] if 'grid_rounds' in batch_config else 2
    interp_func = batch_config['interp_func'] if 'interp_func' in batch_config else 'multiquadric'

    scaler = Scaler()
    x_exist_scaled = scaler.fit_transform(exist_dfc.drop('output', axis=1))
    y_exist = exist_dfc['output']

    krr_model = get_optimal_krr(x_exist_scaled, y_exist, krr_param_grid, grid_rounds, alpha_number)
    initial_score = krr_model.best_score_

    exist_dfc['score_diff'] = 0

    for i, _ in enumerate(exist_dfc.iterrows()):

        exist_minus = exist_dfc.drop(i)

        scaler = Scaler()
        x_minus_scaled = scaler.fit_transform(exist_minus.drop(['output', 'score_diff'], axis=1))
        y_minus = exist_minus['output']

        krr_model = get_optimal_krr(x_minus_scaled, y_minus, krr_param_grid, grid_rounds, alpha_number)
        minus_score = krr_model.best_score_
        score_diff = initial_score - minus_score
        exist_dfc.loc[i, 'score_diff'] = score_diff

    exist_dfc = exist_dfc.drop('output', axis=1)
    cand_dfc = cand_dfc.drop('output', axis=1)

    rbf = Rbf(*[np.array(exist_dfc[column]) for column in list(exist_dfc.columns)], function=interp_func)
    scorediffs = rbf(*[np.array(cand_dfc[column]) for column in list(cand_dfc.columns)])

    return scorediffs


def _looa(cand_df, exist_df, batch_config):
    scorediffs = _leaveoneout(cand_df, exist_df, batch_config)
    scorediffs = np.abs(scorediffs)
    return scorediffs


def _loom(cand_df, exist_df, batch_config):
    scorediffs = -1 * _leaveoneout(cand_df, exist_df, batch_config)
    return scorediffs


def _gugs(cand_df, exist_df, batch_config):  # Gaussian Uncertainty with Grid Search

    alpha_number = batch_config['alpha_number'] if 'alpha_number' in batch_config else 10

    if 'gpr_param_grid' in batch_config:
        gpr_param_grid = batch_config['gpr_param_grid']
    else:
        gpr_param_grid = {
            'kernel': [RationalQuadratic(alpha=0.1, length_scale=0.1)],
            'alpha': np.geomspace(1e-4, 1e4, alpha_number),
            'n_restarts_optimizer': [0, 5, 10]
        }

    Scaler = batch_config['scaler'] if 'scaler' in batch_config else StandardScaler
    grid_rounds = batch_config['grid_rounds'] if 'grid_rounds' in batch_config else 2

    scaler = Scaler()
    x_exist_scaled = scaler.fit_transform(exist_df.drop('output', axis=1))
    y_exist = exist_df['output']

    gpr_model = get_optimal_gpr(x_exist_scaled, y_exist, gpr_param_grid, grid_rounds, alpha_number)

    x_cand_scaled = scaler.transform(cand_df.drop('output', axis=1))
    _,  y_cand_std = gpr_model.best_estimator_.predict(x_cand_scaled, return_std=True)

    return y_cand_std


def _scale_objective(objective_array, batch_config):

    if 'standard' in batch_config:
        if batch_config['standard']:
            objective_array = (objective_array - np.mean(objective_array)) / np.std(objective_array)
            objective_array = (objective_array - np.min(objective_array)) / np.ptp(objective_array)
        else:
            objective_array = (objective_array - np.min(objective_array)) / np.ptp(objective_array)
    else:
        objective_array = (objective_array - np.min(objective_array)) / np.ptp(objective_array)

    return objective_array


def __gaussian_fit(cand_df, exist_df, batch_config):

    scaler = batch_config['scaler'] if 'scaler' in batch_config else StandardScaler()
    gpr = batch_config['gpr'] if 'gpr' in batch_config else GaussianProcessRegressor(kernel=Matern(nu=0.5))

    x_exist = np.array(exist_df.drop('output', axis=1))
    y_exist = np.array(exist_df['output'])
    x_cand = np.array(cand_df.drop('output', axis=1))

    x_exist_norm = scaler.fit_transform(x_exist)
    x_cand_norm = scaler.transform(x_cand)

    gpr.fit(x_exist_norm, y_exist)
    y_cand_mean, y_cand_std = gpr.predict(x_cand_norm, return_std=True)

    return y_cand_mean, y_cand_std



import copy

import numpy as np
from symbal.penalties import invquad_penalty
import pandas as pd
import warnings
import itertools
import sympy

from sklearn.preprocessing import MinMaxScaler, StandardScaler, PolynomialFeatures
from sklearn.linear_model import Lasso
from sklearn.kernel_ridge import KernelRidge
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.model_selection import GridSearchCV
from typing import Union, List
from pysr import PySRRegressor
from scipy.spatial.distance import cdist


def new_penalty(penalty_array, penalty_function, a, b, by_range, batch_size):
    """
    Selects maximum from given values and penalizes area around selection.

    Assumes first column in penalty_array is penalized value, other columns are
    independent variables.

    Returns: independent variables for selected point & new penalty_array w/ penalized values
    """

    max_index = np.nanargmax(penalty_array[:, 0])  # index for largest value
    penalty_array[max_index, 0] = np.nan
    max_pos = penalty_array[max_index, 1:]  # independent variable values for this index

    r_x = np.abs(penalty_array[:, 1:] - max_pos)  # Distance to selected point for each variable
    if by_range:
        s_x = np.ptp(penalty_array[:, 1:], axis=0) / batch_size  # Tune width of penalty by range / batch_size
    else:
        s_x = np.nanstd(penalty_array[:, 1:],
                        axis=0)  # Tune width of penalty by standard deviation of each independent variable
    s_y = np.nanstd(penalty_array[:, 0], axis=0)  # Standard deviation of penalized value

    penalty = penalty_function(a, b, r_x, s_x, s_y)
    penalty_array[:, 0] -= penalty  # subtract penalty

    return max_index, penalty_array


def batch_selection(uncertainty_array, penalty_function=invquad_penalty, a=1, b=1, by_range=False, batch_size=10,
                    **kwargs):

    captured_penalties = pd.DataFrame()
    selected_indices = []
    penalty_array = uncertainty_array

    for i in range(batch_size):

        selected_index, penalty_array = new_penalty(penalty_array, penalty_function, a, b, by_range, batch_size)

        captured_penalties[f'{i}'] = penalty_array[:, 0]
        selected_indices.append(selected_index)

    return selected_indices, captured_penalties


def get_score(input_df, pysr_model):  # TODO - NaN filtering?

    predicted = pysr_model.predict(input_df)
    actual = np.array(input_df['output'])
    score = np.nanmean(np.abs(predicted - actual))

    return score


def get_metrics(pysr_model):

    best_index = np.argmax(pysr_model.equations_['score'])
    equation = pysr_model.equations_.loc[best_index, 'equation']
    loss = pysr_model.equations_.loc[best_index, 'loss']
    score = pysr_model.equations_.loc[best_index, 'score']

    other_equations = pysr_model.equations_.drop(best_index, axis=0)
    loss_other = np.mean(other_equations['loss'])
    score_other = np.mean(other_equations['score'])

    return equation, loss, score, loss_other, score_other


def get_gradient(x_cand, pysr_model, num=None, difference=None):

    if difference is None:
        difference = 1e-8

    overall = copy.deepcopy(x_cand)
    diff_dict = {
        column: pd.concat([x_cand.loc[:, column] + difference, x_cand.drop(column, axis=1)], axis=1)
        for column in x_cand
    }
    for column in x_cand:
        if num is not None:
            overall.loc[:, f'fh__{column}'] = pysr_model.predict(diff_dict[column], num)
            overall.loc[:, f'd__{column}'] = (overall.loc[:, f'fh__{column}'] -
                                              pysr_model.predict(x_cand, num)) / difference
        else:
            overall.loc[:, f'fh__{column}'] = pysr_model.predict(diff_dict[column])
            overall.loc[:, f'd__{column}'] = (overall.loc[:, f'fh__{column}'] -
                                              pysr_model.predict(x_cand)) / difference

    grad_string = 'sqrt('
    for column in overall:
        if 'd__' in column:
            grad_string += f'{column}**2 + '
    grad_string = grad_string.rstrip(' + ') + ')'
    overall['grad'] = overall.eval(grad_string)

    return np.array(overall['grad'])


def get_curvature(x_cand, pysr_model, num=None, difference=None):

    if difference is None:
        difference = 1e-8

    overall = copy.deepcopy(x_cand)
    diff_dict = {
        column: pd.concat([x_cand.loc[:, column] + difference, x_cand.drop(column, axis=1)], axis=1)
        for column in x_cand
    }
    diff2_dict = {
        column: pd.concat([x_cand.loc[:, column] + 2 * difference, x_cand.drop(column, axis=1)], axis=1)
        for column in x_cand
    }
    for column in x_cand:
        if num is not None:
            overall.loc[:, f'fh__{column}'] = pysr_model.predict(diff_dict[column], num)
            overall.loc[:, f'f2h__{column}'] = pysr_model.predict(diff2_dict[column], num)
            overall.loc[:, f'd2__{column}'] = (pysr_model.predict(x_cand, num) - 2 * overall.loc[:, f'fh__{column}'] +
                                               overall.loc[:, f'f2h__{column}']) / difference ** 2
        else:
            overall.loc[:, f'fh__{column}'] = pysr_model.predict(diff_dict[column])
            overall.loc[:, f'f2h__{column}'] = pysr_model.predict(diff2_dict[column])
            overall.loc[:, f'd2__{column}'] = (pysr_model.predict(x_cand) - 2 * overall.loc[:, f'fh__{column}'] +
                                               overall.loc[:, f'f2h__{column}']) / difference ** 2

    lapl_string = ''
    for column in overall:
        if 'd2__' in column:
            lapl_string += f'abs({column}) + '
    lapl_string = lapl_string.rstrip(' + ')
    overall['lapl'] = overall.eval(lapl_string)

    return np.array(overall['lapl'])


def get_all_gradients(x_cand, pysr_model, difference=1e-8):

    gradients = np.empty((len(x_cand), len(pysr_model.equations_['equation'])))

    for j, _ in enumerate(pysr_model.equations_['equation']):
        gradients[:, j] = get_gradient(x_cand, pysr_model, num=j, difference=difference)

    return gradients


def get_uncertainties(x_cand, pysr_model):

    uncertainties = np.empty((len(x_cand), len(pysr_model.equations_['equation'])))
    equation_best = pysr_model.predict(x_cand)

    for j, _ in enumerate(pysr_model.equations_['equation']):
        uncertainties[:, j] = pysr_model.predict(x_cand, j) - equation_best

    return uncertainties


class CustomStrPrinter(sympy.printing.str.StrPrinter):
    def _print_Float(self, expr):
        return f'{expr:.3e}'


def get_equation(column_list: List[str], model: Lasso, scaler: Union[MinMaxScaler, StandardScaler],
                 polynomial: PolynomialFeatures):

    if isinstance(scaler, StandardScaler):
        warnings.warn('Equation identification not yet implemented for StandardScaler.')
        return ''

    if polynomial.interaction_only:
        warnings.warn('Equation identification not yet implemented for interaction_only option.')
        return ''

    variable_list = ['1', *column_list]
    variable_list.extend(list(itertools.combinations_with_replacement(column_list, polynomial.degree)))
    variable_list = np.array(['*'.join(var) if isinstance(var, tuple) else var for var in variable_list])

    lasso_coef = model.coef_
    scaler_range = scaler.data_range_
    scaler_min = scaler.data_min_
    intercept = model.intercept_

    variable_list = variable_list[lasso_coef > 0]
    scaler_range = scaler_range[lasso_coef > 0]
    scaler_min = scaler_min[lasso_coef > 0]
    lasso_coef = lasso_coef[lasso_coef > 0]

    unscaled_coef = lasso_coef * scaler_range
    adj_intercept = intercept + np.sum(lasso_coef * scaler_min)

    equation_terms = [f'{coef:.3e}*{feat}' for coef, feat in zip(unscaled_coef, variable_list)]
    equation = ' + '.join(equation_terms)
    equation += f' + {adj_intercept:.3e}'

    parsed_equation = CustomStrPrinter().doprint(sympy.parsing.sympy_parser.parse_expr(equation))

    return parsed_equation


def get_test(test_config, pysr_model):

    if test_config['test_file']:
        test_df = pd.read_csv(test_config['test_file'], index_col=0)
    elif test_config['test_df']:
        test_df = test_config['test_df']
    else:
        raise RuntimeWarning('Must set either test_file or test_df for test_config.')

    test_df['pred_y'] = pysr_model.predict(test_df)

    variances = test_df.groupby(test_config['groupby'])['pred_y'].var().mean()
    error = np.sqrt(np.mean((test_df[test_config['output_name']] - test_df['pred_y']) ** 2))

    total_err = error + variances

    return total_err


def get_fim(cand_set, pysr_model, single=True, num=None, difference=1e-8):

    if not isinstance(pysr_model, PySRRegressor):
        raise RuntimeWarning('Fischer information matrix (FIM) currently only defined for PySRRegressor.')

    cand_set = cand_set[pysr_model.column_list]

    if single:
        cand_set = np.array([cand_set])
    else:
        cand_set = np.array(cand_set)

    if num is None:
        jax = pysr_model.jax()
    else:
        jax = pysr_model.jax(num)

    parameter_set = np.array([jax['parameters'] for i, _ in enumerate(jax['parameters'])])
    diff_vector = parameter_set.diagonal() * (1 + difference)
    np.fill_diagonal(parameter_set, diff_vector)

    fh = np.zeros((len(cand_set), len(jax['parameters'])))
    f = np.zeros((len(cand_set), len(jax['parameters'])))

    for i, _ in enumerate(cand_set):
        for j, _ in enumerate(jax['parameters']):
            fh[i, j] = jax['callable'](np.array([cand_set[i, :]]), parameter_set[j])[0]
            f[i, j] = jax['callable'](np.array([cand_set[i, :]]), jax['parameters'])[0]

    q = (fh - f) / (parameter_set.diagonal() * difference)
    fim = np.outer(q, q.T)

    return fim


def get_distances(exist_df, x_cand):

    act_y = np.array(exist_df['output'])
    boundary_distance = np.mean(np.abs(act_y))

    x_exist = exist_df.drop('output', axis=1)

    exist_array = np.array(x_exist)
    cand_array = np.array(x_cand)

    exist_norm = (exist_array - np.min(cand_array, axis=0)) / np.ptp(cand_array, axis=0)

    point_distances = cdist(exist_norm, exist_norm)
    point_distance = np.mean(point_distances)

    return boundary_distance, point_distance


def get_class_scores(exist_df, pysr_model):

    y_act = np.array(exist_df['output'])
    act_x = exist_df.drop('output', axis=1)

    y_pred = pysr_model.predict(act_x)
    class_scores = np.abs(np.sign(y_act) - np.sign(y_pred)) / 2
    mod_class_scores = class_scores * np.exp(-np.sqrt(np.abs(y_act) / np.ptp(y_act)))

    class_score = np.mean(class_scores)
    mod_class_score = np.mean(mod_class_scores)

    return class_score, mod_class_score


def get_optimal_krr(x_exist_scaled, y_exist, krr_param_grid, grid_rounds, alpha_number):

    krr_model = GridSearchCV(KernelRidge(), param_grid=krr_param_grid)
    krr_model.fit(x_exist_scaled, y_exist)

    for _ in range(grid_rounds):

        best_alpha = krr_model.best_params_['alpha']
        alpha_space = krr_param_grid['alpha'][1] / krr_param_grid['alpha'][0]
        new_alphas = [alpha for alpha in np.geomspace(best_alpha/alpha_space, best_alpha*alpha_space, alpha_number)]
        krr_param_grid['alpha'] = new_alphas

        krr_model = GridSearchCV(KernelRidge(), param_grid=krr_param_grid)
        krr_model.fit(x_exist_scaled, y_exist)

    return krr_model


def get_optimal_gpr(x_exist_scaled, y_exist, gpr_param_grid, grid_rounds, alpha_number):

    gpr_model = GridSearchCV(GaussianProcessRegressor(), param_grid=gpr_param_grid)
    gpr_model.fit(x_exist_scaled, y_exist)

    for _ in range(grid_rounds):

        best_alpha = gpr_model.best_params_['alpha']
        alpha_space = gpr_param_grid['alpha'][1] / gpr_param_grid['alpha'][0]
        new_alphas = [alpha for alpha in np.geomspace(best_alpha / alpha_space, best_alpha * alpha_space, alpha_number)]
        gpr_param_grid['alpha'] = new_alphas

        gpr_model = GridSearchCV(GaussianProcessRegressor(), param_grid=gpr_param_grid)
        gpr_model.fit(x_exist_scaled, y_exist)

    return gpr_model


def keep_old(old_model, new_model, exist_df, x_cand, batch_config):

    x_exist = exist_df.drop('output', axis=1)
    y_exist = exist_df['output']

    filters = batch_config['filters'] if 'filters' in batch_config else ['mae', 'std']
    std_thres = batch_config['std_thres'] if 'std_thres' in batch_config else 5
    available_filters = ['mae', 'std', 'neg', 'nan', 'inf']

    keep_old_bool = False

    for filter_method in filters:

        if filter_method not in available_filters:
            raise KeyError(f'{filter_method} does not exist as a filtering method. Available methods: '
                           f'{available_filters}')

        past_pred = old_model.predict(x_exist)
        curr_pred = new_model.predict(x_exist)
        past_proj = np.array(old_model.predict(x_cand))
        curr_proj = np.array(new_model.predict(x_cand))

        if filter_method == 'mae':

            y_true = np.array(y_exist)
            past_mae = np.nanmean(np.abs(past_pred - y_true))
            curr_mae = np.nanmean(np.abs(curr_pred - y_true))

            if curr_mae > past_mae:

                keep_old_bool = True
                break

        if filter_method == 'std':

            y_min_thres = np.min([np.mean(y_exist) - std_thres * np.std(y_exist), np.min(y_exist)])
            y_max_thres = np.max([np.mean(y_exist) + std_thres * np.std(y_exist), np.max(y_exist)])

            past_over = past_proj - y_max_thres
            past_over[past_over < 0] = 0
            past_under = y_min_thres - past_proj
            past_under[past_over < 0] = 0
            past_oob = np.sum(past_over + past_under)

            curr_over = curr_proj - y_max_thres
            curr_over[curr_over < 0] = 0
            curr_under = y_min_thres - curr_proj
            curr_under[curr_over < 0] = 0
            curr_oob = np.sum(curr_over + curr_under)

            if curr_oob > past_oob:

                keep_old_bool = True
                break

        if filter_method == 'neg':

            past_proj_copy = copy.deepcopy(past_proj)

            past_proj_copy[past_proj_copy > 0] = 0
            past_oob = np.sum(np.abs(past_proj_copy))

            curr_proj[curr_proj > 0] = 0
            curr_oob = np.sum(np.abs(curr_proj))

            if curr_oob > past_oob:

                keep_old_bool = True
                break

        if filter_method == 'nan':

            past_nan = np.sum(np.isnan(past_proj))
            curr_nan = np.sum(np.isnan(curr_proj))

            if curr_nan > past_nan:

                keep_old_bool = True
                break

        if filter_method == 'inf':

            past_inf = np.sum(np.isinf(past_proj))
            curr_inf = np.sum(np.isinf(curr_proj))

            if curr_inf > past_inf:

                keep_old_bool = True
                break

    return keep_old_bool

